from typing import Callable, Generic, TypeVar

from packyak.asset.asset import AssetNode, TInput, TOutput
from packyak.asset.manifest import Manifest


TManifest = TypeVar("TManifest", bound=Manifest)


class MaterializedAssetNode(Generic[TInput, TOutput], AssetNode[TOutput]):
    def __init__(self, upstream: AssetNode[TInput], f: Callable[[TInput], TOutput]):
        self.upstream = upstream
        self.f = f

    def materialize(self, input: TInput) -> TOutput:
        return self.f(input)


def source(id: str, data: type[TManifest]) -> "SourceAssetNode[TManifest]":
    return SourceAssetNode[TManifest](id, data)


class SourceAssetNode(Generic[TManifest], AssetNode[TManifest]):
    def __init__(self, id: str, type: type[TManifest]):
        self.id = id
        self.type = type

    def table(self):
        def wrapper(
            f: Callable[[TManifest], TOutput],
        ) -> MaterializedAssetNode[TManifest, TOutput]:
            return MaterializedAssetNode(self, f)

        return wrapper



class Resource:
    def __init__(self, resource_id: str):
        self.resource_id = resource_id
        self.bucket_id = self.resource_id
        if resource_id in RESOURCES:
            raise ValueError(f"Resource with id {resource_id} already exists")
        RESOURCES[self.resource_id] = self

    @property
    def resource_type(self) -> str:
        return self.__class__.__name__


RESOURCES: dict[str, Resource] = {}



from typing import Any, Callable, Generic, NamedTuple, TypeVar

Asset = NamedTuple

TInput = TypeVar("TInput", bound=Asset)
TOutput = TypeVar("TOutput", bound=Asset)


class AssetNode(Generic[TOutput]):
    def __init__(self, input: "AssetNode[Any]") -> None:
        self.input = input


class TableAssetNode(AssetNode[TOutput]):
    pass


def asset():
    def wrapper(func: Callable[..., Asset]):
        pass

    return wrapper



def file_path_to_parent_package(file_path: str) -> str:
    components = file_path.split("/")
    package_name = ".".join(components[:-1])
    return package_name


def file_path_to_module_name(file_path: str) -> str:
    file_path = file_path.rsplit(".", 1)[0]
    components = file_path.split("/")
    package_name = ".".join(components)
    return package_name



from typing import Callable, TypeVar, ParamSpec
from packyak.runtime.runnable import Runnable
from packyak.util.fqn import get_fully_qualified_name


from packyak.spec import DependencyGroup
from packyak.duration import Duration

Params = ParamSpec("Params")
Return = TypeVar("Return")


class LambdaFunction(Runnable[Params, Return]):
    def __init__(
        self,
        file_name: str,
        handler: Callable[Params, Return],
        function_id: str | None = None,
        memory: int | None = None,
        timeout: Duration | None = None,
        with_: DependencyGroup | None = None,
        without: DependencyGroup | None = None,
        dev: bool | None = None,
        all_extras: bool | None = None,
        without_hashes: bool | None = None,
        without_urls: bool | None = None,
    ) -> None:
        super().__init__(
            resource_id=function_id or get_fully_qualified_name(handler),
            handler=handler,
            file_name=file_name,
            with_=with_,
            without=without,
            dev=dev,
            all_extras=all_extras,
            without_hashes=without_hashes,
            without_urls=without_urls,
        )

        self.memory = memory
        self.timeout = timeout


def function(
    *,
    function_id: str | None = None,
    memory: int | None = None,
    timeout: Duration | None = None,
    file_name: str | None = None,
    with_: DependencyGroup = None,
    without: DependencyGroup = None,
    # deprecated, use with_ and without
    dev: bool | None = None,
    all_extras: bool | None = None,
    without_hashes: bool | None = None,
    without_urls: bool | None = None,
):
    def decorator(handler: Callable[Params, Return]) -> LambdaFunction[Params, Return]:
        _function_id = (
            function_id
            if function_id is not None
            else get_fully_qualified_name(handler)
        )

        func = LambdaFunction(
            function_id=_function_id,
            memory=memory,
            timeout=timeout,
            handler=handler,
            file_name=file_name or handler.__code__.co_filename,
            with_=with_,
            without=without,
            dev=dev,
            all_extras=all_extras,
            without_hashes=without_hashes,
            without_urls=without_urls,
        )

        return func

    return decorator



from __future__ import annotations

from typing import List, Literal, TypeVar
from enum import Enum
from pydantic import BaseModel

BucketSubscriptionScope = Literal["create"] | Literal["delete"]


BucketIntegrationScope = (
    Literal["get"] | Literal["list"] | Literal["put"] | Literal["delete"]
)


Item = TypeVar("Item")

NonEmptyList = list[Item]

DependencyGroup = NonEmptyList[str] | str | None


class BindingSpec(BaseModel):
    resource_type: str
    resource_id: str
    scopes: list[str]
    props: dict[str, str] | None


class BucketBindingSpec(BindingSpec):
    selector: str | None


class PythonPoetryArgs(BaseModel):
    with_: DependencyGroup | None = None
    without: DependencyGroup | None = None
    dev: bool | None = None
    all_extras: bool | None = None
    without_hashes: bool | None = None
    without_urls: bool | None = None


# defines the bindings within a Python module (file)
class ModuleSpec(PythonPoetryArgs):
    file_name: str
    bindings: list[BindingSpec] | None


class FunctionSpec(PythonPoetryArgs):
    function_id: str
    file_name: str
    bindings: list[BindingSpec] | None


class JobSpec(PythonPoetryArgs):
    job_id: str
    file_name: str
    bindings: list[BindingSpec] | None


class BucketSubscriptionSpec(BaseModel):
    scopes: List[BucketSubscriptionScope]
    function_id: str


class BucketSpec(BaseModel):
    bucket_id: str
    subscriptions: list[BucketSubscriptionSpec]


class QueueSubscriptionSpec(BaseModel):
    function_id: str


class QueueSpec(BaseModel):
    queue_id: str
    fifo: bool
    subscriptions: list[QueueSubscriptionSpec]


class ClusterSpec(BaseModel):
    cluster_id: str
    # bindings: list[BindingSpec] | None


class PackyakSpec(BaseModel):
    modules: list[ModuleSpec]
    buckets: list[BucketSpec]
    queues: list[QueueSpec]
    clusters: list[ClusterSpec]
    functions: list[FunctionSpec]
    jobs: list[JobSpec]



from typing import Callable, ParamSpec, TypeVar, Generic
from packyak.resource import Resource
from packyak.util.fqn import get_fully_qualified_name
from packyak.spec import DependencyGroup

Params = ParamSpec("Params")
Return = TypeVar("Return")


class Runnable(Resource, Generic[Params, Return]):
    def __init__(
        self,
        handler: Callable[Params, Return],
        resource_id: str | None,
        file_name: str | None,
        with_: DependencyGroup | None,
        without: DependencyGroup | None,
        dev: bool | None,
        all_extras: bool | None,
        without_hashes: bool | None,
        without_urls: bool | None,
    ):
        super().__init__(
            resource_id=resource_id or get_fully_qualified_name(handler),
        )
        self.file_name = file_name or handler.__code__.co_filename
        self.handler = handler
        self.with_ = with_
        self.without = without
        self.dev = dev
        self.all_extras = all_extras
        self.without_hashes = without_hashes
        self.without_urls = without_urls



from typing import TYPE_CHECKING, Callable
from enum import Enum
from packyak.asset.asset import AssetNode, TInput, TOutput, TableAssetNode

if TYPE_CHECKING:
    import pyspark.sql as sql
    import pyspark.pandas as ps


class TableFormat(Enum):
    AVRO = "avro"
    CSV = "csv"
    DELTA = "delta"
    JSON = "json"
    ORC = "orc"
    PARQUET = "parquet"


class Namespace:
    def __init__(self, name: str, parent: "Namespace | None" = None):
        self.name = name
        self.parent = parent
        self.children: dict[str, Namespace] = {}

    def __getattr__(self, name: str) -> "Namespace":
        if name not in self.children:
            self.children[name] = Namespace(name, self)
        return self.children[name]

    def table(
        self,
        *,
        input: AssetNode[TInput],
        output: type[TOutput] | None = None,
        format: TableFormat = TableFormat.PARQUET,
    ):
        def create_table_node(
            f: Callable[
                [ps.DataFrame[TInput]],
                ps.DataFrame[TOutput] | sql.DataFrame,
            ]
            | Callable[
                [sql.DataFrame],
                ps.DataFrame[TOutput] | sql.DataFrame,
            ],
        ) -> TableAssetNode[TOutput]:
            return TableAssetNode[TOutput](input=input)

        return create_table_node


class DB(Namespace):
    def __init__(self, name: str):
        super().__init__(name)
        if name in dbs:
            raise ValueError(f"DB {name} already exists")
        dbs[name] = self


dbs: dict[str, DB] = {}



from packyak.cli.cli import cli
from packyak.cli.list import list
from packyak.cli.materialize import materialize
from packyak.cli.new import new
from packyak.cli.run import run
from packyak.cli.ssh import ssh
from packyak.cli.synth import synth

__all__ = ["cli", "new", "ssh", "synth", "run", "materialize", "list"]

cli()



from _ast import AsyncFunctionDef, FunctionDef
import ast
from functools import partial
import inspect
from textwrap import dedent
from typing import Any, Callable, cast

from packyak.runtime.binding import Binding
from packyak.runtime.function import LambdaFunction
from packyak.runtime.job import Job

from packyak.runtime.integration import Integration, is_integration
from packyak.resource import Resource
from packyak.runtime.runnable import Runnable
from packyak.synth.call import Call
from packyak.synth.loaded_module import LoadedModule

Func = Callable[..., Any]


def bind(func: Runnable[Any, Any] | LoadedModule) -> list[Binding]:
    calls = (
        analyze_loaded_module(func)
        if isinstance(func, LoadedModule)
        else analyze_function(func.handler)
    )

    return [
        Binding(func, call.obj, call.func.scopes, call.metadata)
        for call in calls
        if call.obj is not None
    ]


def analyze_loaded_module(module: LoadedModule) -> list[Call]:
    calls: list[Call] = []

    attr_dict = module.vars
    seen = set[Any]()

    def should_skip(node: ast.AST) -> bool:
        if isinstance(node, ast.FunctionDef) or isinstance(node, ast.AsyncFunctionDef):
            if node.name in attr_dict:
                func = attr_dict[node.name]
                if isinstance(func, LambdaFunction):
                    # Hack: terminate analysis if we come across a LambdaFunction
                    return True
        return False

    def eval_expr(node: ast.AST | None) -> Any:
        if isinstance(node, ast.Name):
            return eval(node.id, attr_dict)
        elif isinstance(node, ast.Constant):
            return node.value
        elif isinstance(node, ast.Attribute):
            obj = eval_expr(node.value)
            return getattr(obj, node.attr)
        else:
            raise Exception(f"Unsupported node type: {type(node)}")

    class Visitor(ast.NodeVisitor):
        def visit_FunctionDef(self, node: FunctionDef) -> Any:
            if should_skip(node):
                return node
            return super().generic_visit(node)

        def visit_AsyncFunctionDef(self, node: AsyncFunctionDef) -> Any:
            if should_skip(node):
                return node
            return super().generic_visit(node)

        def visit_Call(self, node: ast.Call):
            try:
                func = eval_expr(node.func)
                if isinstance(func, type) and issubclass(func, Resource):
                    return node
                else:
                    obj = (
                        eval_expr(node.func.value)
                        if isinstance(node.func, ast.Attribute)
                        else func
                    )
                    calls.extend(analyze_function(func, seen, obj=obj))
            except Exception:
                return super().generic_visit(node)

    visitor = Visitor()
    visitor.visit(module.ast)
    return calls


def analyze_function(
    func: Func, seen: set[Func] = set(), *, obj: object | None = None
) -> list[Call]:
    if func in seen:
        return []

    seen.add(func)

    if is_integration(func) and obj is not None:
        # this is an Integration
        integration = cast(Integration[Any, Any], func)
        metadata = {}
        if integration.metadata is not None:
            for k, v in integration.metadata.items():
                if isinstance(v, property):
                    metadata[k] = getattr(obj, v.fget.__name__)
                else:
                    metadata[k] = getattr(obj, v)

        return [
            Call(
                func=func,  # type: ignore
                obj=obj,
                metadata=metadata,
            )
        ]
    elif isinstance(func, partial):
        print("partial")
        return analyze_function(func.func, seen, obj=obj)

    lexical_scope = get_lexical_scope(func)

    def lookup(expr: str):
        return eval(expr, lexical_scope)

    # Get the source of the function
    try:
        src = dedent(inspect.getsource(func))
    except TypeError:
        # TODO: Handle builtins
        return []

    # Parse the source into an AST
    module = ast.parse(src)

    # Custom visitor class to find function calls
    class CallVisitor(ast.NodeVisitor):
        calls: list[Call]

        def __init__(self):
            self.calls = []

        def visit_Call(self, node: ast.Call):
            try:
                if isinstance(node.func, ast.Name):
                    self.calls.extend(analyze_function(lookup(node.func.id), seen))
                elif isinstance(node.func, ast.Attribute) and isinstance(
                    node.func.value, ast.Name
                ):
                    calls = analyze_function(
                        lookup(f"{node.func.value.id}.{node.func.attr}"),
                        seen,
                        obj=lookup(node.func.value.id),
                    )
                    self.calls.extend(calls)

                # elif isinstance(node.func, ast.Subscript):
                #     call = Call(
                #         lookup(f"{node.func.value.id}[{node.func.slice.value.s!r}]"),
                #         lookup(f"{node.func.value.id}"),
                #     )
                #     self.calls.append(call)
            except NameError:
                pass

    visitor = CallVisitor()
    visitor.visit(module)
    return visitor.calls


# Get the lexical scope of a function
def get_lexical_scope(func: Func) -> dict[str, Any]:
    if isinstance(func, partial):
        return get_lexical_scope(func.func)

    globals = func.__globals__ if hasattr(func, "__globals__") else {}
    closure_vars = {}
    if hasattr(func, "__closure__") and func.__closure__ is not None:
        closure_var_names = func.__code__.co_freevars
        closure_vars = {
            name: cell.cell_contents
            for name, cell in zip(closure_var_names, func.__closure__)
        }

    lexical_scope = {**globals, **closure_vars}

    return lexical_scope



import click
from packyak.cli.cli import cli
from packyak.util.git import get_git_branch


@cli.command()
@click.argument("root_dir", type=str)
@click.option(
    "--branch",
    type=str,
    help="Name of the data branch to materialize. Default to the branch of the current git repository.",
)
@click.option(
    "--profile",
    type=str,
    help="AWS CLI profile to use when authenticating to SSM",
)
@click.option(
    "--asset",
    multiple=True,
    type=str,
    help="List of assets to materialize. Can be specified multiple times. Defaults to all.",
)
def run(
    root_dir: str,
    branch: str = get_git_branch(),
    profile: str | None = None,
    assets: list[str] = [],
):
    pass



from typing import Any
import click
import subprocess
import time
import boto3
import os
import re
import collections

from packyak.cli.cli import cli
from packyak.util.emr import EMR


@cli.command()
@click.argument(
    "instance-id",
    type=str,
    # prompt="What is the ID of the EC2 Instance or EMR Cluster you wish to tunnel to?",
    # help="The EC2 Instance ID of the instance you wish to tunnel to",
)
@click.option(
    "--ssh-key",
    default="~/.ssh/id_rsa",
    help="Path to the the SSH public key to copy to the remote host. Default: ~/.ssh/id_rsa.pub",
)
@click.option(
    "-L",
    "port_forwards",
    # type=(str | list[str]),
    multiple=True,
    help="Ports to forward. Default: 9000:localhost:22",
)
@click.option("-v", "--verbose", type=str, is_flag=True)
@click.option(
    "--profile", type=str, help="AWS CLI profile to use when authenticating to SSM"
)
@click.option("--user", type=str, help="POSIX username to log in as. Defaults to root.")
def ssh(
    instance_id: str,
    ssh_key: str = "~/.ssh/id_rsa",
    # todo: what port forwarding to i need for VS Code Remote SSH
    port_forwards: list[str] = ["9001:localhost:22"],
    verbose: bool = False,
    profile: str | None = None,
    user: str | None = None,
):
    """
    Establishes a secure tunnel to an EC2 instance.

    instance-id: The EC2 Instance ID of the instance you wish to tunnel to.

    --ssh-public-key: Path to the SSH public key to copy to the remote host. Default: ~/.ssh/id_rsa.pub

    -L: Ports to forward. Default: 9001:localhost:22
    """

    if profile is not None:
        os.environ["AWS_PROFILE"] = profile

    def log(message: str):
        if verbose:
            print(message)

    def resolve_primary_node(instance_id: str) -> str:
        emr = EMR()

        return emr.get_primary_node_instance_id(cluster_id=instance_id)

    if instance_id.startswith("j"):
        log(
            f"Instance ID {instance_id} looks like an EMR cluster ID. Resolving the Primary node."
        )
        instance_id = resolve_primary_node(instance_id)
        log(f"Resolved primary node: {instance_id}")

    if user is None:
        sts_client = boto3.client("sts")
        caller_identity = sts_client.get_caller_identity()
        user_id = caller_identity.get("UserId", "")
        user = user_id.split(":")[1] if ":" in user_id else "root"
    log(f"Logging in as {user}")

    log(f"{time.ctime()} sm-connect-ssh-proxy: Connecting to: {instance_id}")
    # print(f"{time.ctime()} sm-connect-ssh-proxy: Extra args: {extra_ssh_args}")

    ssm_client = boto3.client("ssm")
    instance_info = ssm_client.describe_instance_information(
        Filters=[{"Key": "InstanceIds", "Values": [instance_id]}]
    )
    if not instance_info["InstanceInformationList"]:
        log("No instance information found.")
        instance_status = "Offline"
    else:
        instance_status = instance_info["InstanceInformationList"][0]["PingStatus"]
    log(f"Instance status: {instance_status}")

    if instance_status != "Online":
        log("Error: Instance is offline.")
        return

    with open(os.path.expanduser(f"{ssh_key}.pub"), "r") as file:
        ssh_pub_key = file.read().strip()

    current_region = boto3.session.Session().region_name
    log(f"Will use AWS Region: {current_region}")

    aws_cli_version = subprocess.check_output(
        ["aws", "--version"], stderr=subprocess.STDOUT
    ).decode()
    log(f"AWS CLI version (should be v2): {aws_cli_version}")

    ssh_dir = "/root/.ssh/" if user == "root" else f"/home/{user}/.ssh/"

    commands = [
        # f'echo "{ssh_pub_key}" > /etc/ssh/authorized_keys',
        f"mkdir -p {ssh_dir}",
        f'echo "{ssh_pub_key}" > {ssh_dir}/authorized_keys',
    ]
    log(f"Sending command to instance: {commands}")
    send_command_response = ssm_client.send_command(
        InstanceIds=[instance_id],
        DocumentName="AWS-RunShellScript",
        Comment="Copy public key for SSH helper",
        Parameters={"commands": commands},
        TimeoutSeconds=30,
    )
    command_id = send_command_response["Command"]["CommandId"]
    log(f"Got command ID: {command_id}")

    # time.sleep(5)  # Wait a bit to prevent InvocationDoesNotExist error

    for _ in range(15):
        try:
            output = ssm_client.get_command_invocation(
                CommandId=command_id, InstanceId=instance_id
            )
            log(f"Command status: {output['Status']}")
            if output["Status"] not in ["Pending", "InProgress"]:
                log(f"Command output: {output.get('StandardOutputContent', '')}")
                if output.get("StandardErrorContent"):
                    log(f"Command error: {output['StandardErrorContent']}")
                break
        except ssm_client.exceptions.InvocationDoesNotExist:
            pass
        time.sleep(0.1)

    if output["Status"] != "Success":  # type: ignore
        log("Error: Command didn't finish successfully in time")
        return

    log(f"{time.ctime()} sm-connect-ssh-proxy: Starting SSH over SSM proxy")

    proxy_command = (
        "aws ssm start-session "
        + "--reason 'PackYak SSH' "
        + f"--region '{current_region}' "
        + f"--target '{instance_id}' "
        + "--document-name AWS-StartSSHSession "
        + "--parameters portNumber=%p"
    )

    # Start with the base SSH command
    command = ["ssh", "-4"]

    # Add each SSH option prefixed by '-o'
    for option in [
        f"User={user}",
        f"IdentityFile={ssh_key}",
        "IdentitiesOnly=yes",
        f"ProxyCommand={proxy_command}",
        "ServerAliveInterval=15",
        "ServerAliveCountMax=3",
        "PasswordAuthentication=no",
        "StrictHostKeyChecking=no",
        "UserKnownHostsFile=/dev/null",
    ]:
        command.extend(["-o", option])

    # Add each port forward prefixed by '-L'
    for port in port_forwards:
        command.extend(["-L", port])

    # Finally, add the instance ID
    command.append(instance_id)

    log(" ".join(command))  # Debugging: log the command

    # Execute the command
    process = subprocess.run(
        command,
        check=True,
    )
    return process.stdout, process.stderr


class Host:
    def __init__(self, name: str, props: dict[str, Any]):
        self.name = name
        self.props = props


def parse_ssh_config(file_path: str = "~/.ssh/config") -> list[Host]:
    """Parse the SSH config file and return a list of Host objects with kwargs for each property."""
    hosts = []

    def parse_key_value(line: str) -> tuple[str, str] | None:
        split = re.split(r"\s+", line.strip(), maxsplit=1)
        if len(split) == 2:
            return split[0], split[1]
        else:
            return None

    with open(file_path, "r") as f:
        lines = f.readlines()

        queue_lines = collections.deque(lines)

        hosts: list[Host] = []

        while queue_lines:
            line = queue_lines.popleft()

            if line.startswith("Host "):
                parsed_host = parse_key_value(line)
                if parsed_host is None:
                    raise ValueError(
                        "Failed to parse host name from SSH config line:\n{line}"
                    )
                host = Host(name=parsed_host[1], props={})
                hosts.append(host)

                # now collect all of its key-value props
                while queue_lines:
                    # Peek at the next line without removing it from the queue
                    line = queue_lines[0]
                    if line.startswith("Host "):
                        # have found a new host, we're done
                        break
                    else:
                        key_value = parse_key_value(line)
                        if key_value is not None:
                            host.props[key_value[0]] = key_value[1]

    return hosts



from typing import Literal

TimeUnit = (
    Literal["second"]
    | Literal["seconds"]
    | Literal["minute"]
    | Literal["minutes"]
    | Literal["hour"]
    | Literal["hours"]
    | Literal["day"]
    | Literal["days"]
)


class Duration:
    def __init__(self, amount: int, unit: TimeUnit) -> None:
        self.amount = amount
        self.unit = unit


def duration(amount: int, unit: TimeUnit) -> Duration:
    return Duration(amount, unit)



from typing import TypeVar, Callable

T = TypeVar("T")


def memoize(fn: Callable[[], T]) -> Callable[[], T]:
    cache = None

    def wrapper():
        nonlocal cache
        if cache is None:
            cache = fn()
        return cache

    return wrapper



from typing import Any

from packyak.runtime.function import LambdaFunction
from packyak.runtime.job import Job
from packyak.resource import Resource
from packyak.runtime.runnable import Runnable
from packyak.spec import BindingSpec
from packyak.synth.loaded_module import LoadedModule


BindingTarget = Runnable[Any, Any] | LoadedModule


class Binding:
    def __init__(
        self,
        function: BindingTarget,
        resource: Resource,
        scopes: list[str],
        metadata: dict[str, Any] | None = None,
    ) -> None:
        self.resource = resource
        self.scopes = scopes
        self.function = function
        self.metadata = metadata

    def to_binding_spec(self) -> BindingSpec:
        return BindingSpec(
            resource_type=self.resource.resource_type,
            resource_id=self.resource.resource_id,
            scopes=self.scopes,
            props=self.metadata,
        )



from packyak.spec import ResourceType
from packyak.resource import Resource


class Pool(Resource):
    def __init__(self, pool_id: str):
        super().__init__(resource_id=pool_id, resource_type=ResourceType.Pool)
        self.name = pool_id



import click
from packyak.cli.cli import cli
from packyak.util.git import get_git_branch
from packyak.synth import synth as _synth
import os
import sys
import aiofiles as aio
import asyncio


@click.option(
    "-c",
    "--config",
    type=str,
    help="Path to the packyak.config.py file.",
    default="./packyak.config.py",
)
@click.option(
    "--branch",
    type=str,
    help="Name of the data branch to materialize. Default to the branch of the current git repository.",
    default=get_git_branch(),
)
@click.option(
    "--profile",
    type=str,
    help="AWS CLI profile to use when authenticating to SSM",
)
@cli.command()
def synth(
    config: str,
    branch: str,
    profile: str | None,
):
    import importlib
    import importlib.util

    # Add the root directory to the path so we can import relative to the user's code
    sys.path.append(os.path.dirname(os.path.abspath(config)))

    config_module_spec = importlib.util.spec_from_file_location(
        "packyak.config", config
    )
    if config_module_spec is None:
        raise ImportError(
            f"Could not find module {config}. Please ensure the file exists and is a valid Python module."
        )
    if config_module_spec.loader is not None:
        config_module = importlib.util.module_from_spec(config_module_spec)
        config_module_spec.loader.exec_module(config_module)
    else:
        raise ImportError(
            f"Could not load module {config}. Please ensure the loader is available."
        )
    asyncio.run(async_synth())


async def async_synth():
    # Your async code here
    packyak_spec = await _synth()  # Assuming _synth can be an async function
    os.makedirs(".packyak", exist_ok=True)
    async with aio.open(os.path.join(".packyak", "manifest.json"), "w") as f:
        await f.write(packyak_spec.model_dump_json(indent=2, exclude_none=True))



from typing import (
    Any,
    Callable,
    Generic,
    Protocol,
    runtime_checkable,
    TypeVar,
    List,
    ParamSpec,
)

P = ParamSpec("P")
R = TypeVar("R", covariant=True)


@runtime_checkable
class Integration(Protocol, Generic[P, R]):
    scopes: List[str]
    metadata: dict[str, Any] | None

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        ...


def is_integration(func: Any) -> bool:
    return hasattr(func, "scopes") and hasattr(func, "metadata")


def integration(*scopes: str, **kwargs: property):
    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        setattr(func, "scopes", scopes)
        setattr(func, "metadata", kwargs if kwargs and len(kwargs) > 0 else None)
        return func

    return decorator



import ast
import types


class LoadedModule:
    def __init__(
        self,
        module: types.ModuleType,
        ast: ast.Module,
        module_name: str,
        file_name: str,
    ):
        self.module = module
        self.ast = ast
        self.module_name = module_name
        self.file_name = file_name
        self.vars = vars(module)



import ast
from importlib import import_module
import os
import types
from typing import Any, TypeVar
import aiofiles

from packyak.storage.bucket import Bucket
from packyak.runtime.cluster import Cluster
from packyak.runtime.function import LambdaFunction
from packyak.runtime.job import Job
from packyak.streaming.queue import Queue
from packyak.resource import RESOURCES, Resource
from packyak.runtime.runnable import Runnable
from packyak.spec import (
    BucketSpec,
    BucketSubscriptionSpec,
    ClusterSpec,
    FunctionSpec,
    JobSpec,
    ModuleSpec,
    PackyakSpec,
    QueueSpec,
    QueueSubscriptionSpec,
)
from packyak.synth.analyze import bind
from packyak.synth.file_utils import file_path_to_module_name
from packyak.synth.loaded_module import LoadedModule


async def synth(root_dir: str | None = None) -> PackyakSpec:
    def visit(resource: Resource):
        if resource in seen:
            return
        seen.add(resource)

        def get_bindings(resource: Runnable[Any, Any]):
            bindings = bind(resource)
            return (
                [binding.to_binding_spec() for binding in bind(resource)]
                if len(bindings) > 0
                else None
            )

        if isinstance(resource, Bucket):
            buckets.append(
                BucketSpec(
                    bucket_id=resource.resource_id,
                    subscriptions=[
                        BucketSubscriptionSpec(
                            scopes=sub.scopes,
                            function_id=sub.function.resource_id,
                        )
                        for sub in resource.subscriptions
                    ],
                )
            )
        elif isinstance(resource, Queue):
            queues.append(
                QueueSpec(
                    queue_id=resource.resource_id,
                    fifo=resource.fifo,
                    subscriptions=[
                        QueueSubscriptionSpec(function_id=sub.function.resource_id)
                        for sub in resource.subscriptions
                    ],
                )
            )
        elif isinstance(resource, LambdaFunction):
            functions.append(
                FunctionSpec(
                    function_id=resource.resource_id,
                    file_name=resource.file_name,
                    bindings=get_bindings(resource),
                    with_=resource.with_,
                    without=resource.without,
                    dev=resource.dev,
                    all_extras=resource.all_extras,
                    without_hashes=resource.without_hashes,
                    without_urls=resource.without_urls,
                )
            )
        elif isinstance(resource, Job):
            jobs.append(
                JobSpec(
                    job_id=resource.resource_id,
                    file_name=resource.file_name,
                    bindings=get_bindings(resource),
                    with_=resource.with_,
                    without=resource.without,
                    dev=resource.dev,
                    all_extras=resource.all_extras,
                    without_hashes=resource.without_hashes,
                    without_urls=resource.without_urls,
                )
            )
        elif isinstance(resource, Cluster):
            clusters.append(ClusterSpec(cluster_id=resource.cluster_id))

    modules: list[ModuleSpec] = []
    functions: list[FunctionSpec] = []
    buckets: list[BucketSpec] = []
    queues: list[QueueSpec] = []
    clusters: list[ClusterSpec] = []
    jobs: list[JobSpec] = []
    seen = set[Any]()

    if root_dir is not None:
        for root, _, files in os.walk(root_dir):
            for file in files:
                if file.endswith(".py"):
                    print(root_dir, file)
                    file_path = os.path.join(root, file)
                    absolute_file_path = os.path.abspath(file_path)
                    async with aiofiles.open(file_path, mode="r") as f:
                        module_ast = ast.parse(await f.read())
                    module_name = file_path_to_module_name(file_path)
                    module: types.ModuleType = import_module(module_name)

                    loaded_module = LoadedModule(
                        module, module_ast, module_name, file_path
                    )

                    bindings = bind(loaded_module)
                    if len(bindings) > 0:
                        modules.append(
                            ModuleSpec(
                                file_name=absolute_file_path,
                                bindings=[
                                    binding.to_binding_spec() for binding in bindings
                                ],
                            )
                        )

    for resource in find_all_resources():
        visit(resource)

    packyak_spec = PackyakSpec(
        modules=modules,
        buckets=buckets,
        queues=queues,
        functions=functions,
        clusters=clusters,
        jobs=jobs,
    )
    return packyak_spec


def lookup_function(function_id: str) -> LambdaFunction[Any, Any]:
    if function_id not in RESOURCES:
        raise Exception(f"Lambda Function {function_id} does not exist")
    resource = RESOURCES[function_id]
    if isinstance(resource, LambdaFunction):
        return resource
    raise Exception(f"Resource {function_id} is not a Lambda Function")


def find_all_resources() -> list[Resource]:
    return list(RESOURCES.values())


T = TypeVar("T", bound=Resource)


def find_resource(t: type[T]) -> list[T]:
    return [resource for resource in RESOURCES.values() if isinstance(resource, t)]



import os

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyspark.sql import SparkSession


def init_session(cwd: str = os.getcwd(), venv: str | None = None) -> SparkSession:
    return session_builder(cwd, venv).getOrCreate()


def session_builder(
    cwd: str = os.getcwd(), venv: str | None = None
) -> SparkSession.Builder:
    import findspark

    findspark.init()

    venv = venv if venv is not None else os.path.join(cwd, ".venv", "bin", "python")

    builder: SparkSession.Builder = (
        SparkSession.builder.master("yarn")  # type: ignore - not sure why builder is of type classproperty instead of SparkSession.builder
        .config("spark.pyspark.python", venv)
        .config("spark.pyspark.driver.python", venv)
        .config("spark.pyspark.virtualenv.enabled", "false")
    )
    return builder



from typing import Any
from packyak.runtime.integration import Integration


class Call:
    def __init__(
        self,
        func: Integration[Any, Any],
        *,
        obj: Any = None,
        metadata: dict[str, property] | None = None,
    ):
        self.func = func
        self.obj = obj
        self.metadata = metadata



from typing import Any, TypeVar, AsyncContextManager

# Define a type variable for the casted type
T = TypeVar("T")


class TypedResource(AsyncContextManager[T]):
    def __init__(self, untyped_resource: AsyncContextManager[Any]):
        self.untyped_resource = untyped_resource
        self.typed_resource: T | None = None  # To be set in __aenter__

    async def __aenter__(self) -> T:
        # Proxy the __aenter__ call to the untyped resource, then cast it
        raw_resource = await self.untyped_resource.__aenter__()
        self.typed_resource = self.cast_to_type(raw_resource)
        return self.typed_resource

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):
        # Proxy the __aexit__ call to the untyped resource
        await self.untyped_resource.__aexit__(exc_type, exc_val, exc_tb)

    def cast_to_type(self, resource: Any) -> T:
        # default implementation is type-only
        return resource



import os
import click
from packyak.cli.cli import cli
from packyak.util.git import get_git_branch
from packyak.synth import synth


@cli.command()
@click.argument(
    "assets",
    nargs=-1,
    type=list[str],
)
@click.option(
    "--root",
    type=str,
    help="The working directory of the PackYak project. Defaults to CWD.",
)
@click.option(
    "--branch",
    type=str,
    help="Name of the data branch to materialize. Default to the branch of the current git repository.",
)
@click.option(
    "--profile",
    type=str,
    help="AWS CLI profile to use when authenticating to SSM",
)
def materialize(
    assets: list[str] = [],
    root: str = os.getcwd(),
    branch: str = get_git_branch(),
    profile: str | None = None,
):
    spec = synth(root)

    pass



from packyak.runtime.function import function
from packyak.runtime.cluster import Cluster, Engine
from packyak.runtime.job import Job
from packyak.duration import Duration, duration, TimeUnit
from packyak.scheduling.every import every
from packyak.scheduling.cron import cron
from packyak.storage.bucket import Bucket
from packyak.streaming.queue import Queue, Message, ReceivedMessagesEvent
from packyak.synth.synth import synth
from packyak.asset.asset import Asset, asset
from packyak.asset.manifest import Manifest
from packyak.asset.namespace import DB, Namespace
from packyak.asset.partition_key import PartitionKey
from packyak.asset.source import source

__all__ = [
    "asset",
    "Asset",
    "Bucket",
    "Cluster",
    "cron",
    "DB",
    "duration",
    "Duration",
    "Engine",
    "every",
    "function",
    "Job",
    "Manifest",
    "Message",
    "Namespace",
    "PartitionKey",
    "Queue",
    "ReceivedMessagesEvent",
    "source",
    "synth",
    "TimeUnit",
]



import os
from datetime import datetime
from typing import (
    IO,
    Any,
    Callable,
    Generic,
    Optional,
    TypeVar,
    cast,
    List,
    Tuple,
)

from pydantic import BaseModel
from packyak.util.fqn import get_fully_qualified_name
from packyak.util.memoize import memoize
from packyak.util.typed_resource import TypedResource
from packyak.storage.folder import Folder
from packyak.runtime.function import LambdaFunction, function
from packyak.runtime.integration import integration
from packyak.spec import BucketSubscriptionScope, DependencyGroup
from packyak.resource import Resource

import aioboto3
import boto3
from botocore.response import StreamingBody
from types_aiobotocore_s3.client import S3Client
from types_aiobotocore_s3.type_defs import ListObjectsV2OutputTypeDef

from pyspark import SparkContext, RDD


s3 = memoize(lambda: boto3.client("s3"))
session = memoize(lambda: aioboto3.Session())
aio_s3 = memoize(lambda: TypedResource[S3Client](session().client("s3")))  # type: ignore


class ObjectRef:
    def __init__(self, bucket: "Bucket", key: str, etag: str):
        self.bucket = bucket
        self.key = key
        self.etag = etag
        self.created_at = datetime.now()

    def __hash__(self):
        return hash((self.bucket.bucket_id, self.key, self.etag, self.created_at))


Body = TypeVar("Body", bound=Any)


class Object(Generic[Body], ObjectRef):
    def __init__(self, bucket: "Bucket", key: str, body: Body, etag: str):
        super().__init__(bucket, key, etag)
        self.body = body


class ListObjectsResponse:
    def __init__(
        self, bucket: "Bucket", objects: list[ObjectRef], next_token: str | None
    ):
        self.bucket = bucket
        self.objects = objects
        self.next_token = next_token

    def __aiter__(self):
        return self

    async def __anext__(self):
        pass


class BucketSubscription:
    def __init__(
        self,
        bucket: "Bucket",
        scopes: List[BucketSubscriptionScope],
        function: LambdaFunction[Any, Any],
        prefix: str | None = None,
    ):
        self.bucket = bucket
        self.scopes = scopes
        self.function = function
        self.prefix = prefix


Blob = str | bytes | IO[Any] | StreamingBody


class Bucket(Resource):
    subscriptions: list[BucketSubscription]

    sc: SparkContext

    def __init__(self, bucket_id: str):
        super().__init__(bucket_id)
        self.subscriptions = []

    class Event(BaseModel):
        key: str
        etag: str

    class ObjectCreatedEvent(Event):
        size: int | None

    class ObjectDeletedEvent(Event):
        pass

    def __str__(self) -> str:
        return f"s3://{self.bucket_name}/"

    # Overload the '/' operator
    def __truediv__(self, name: str) -> Folder:
        return self.folder(name)

    def folder(self, name: str) -> Folder:
        _folder: Optional[Folder] = None
        for path_component in name.split("/"):
            if _folder is None:
                _folder = Folder(self, path_component)
            else:
                _folder = _folder / path_component
        return cast(Folder, _folder)

    @property
    def bucket_name(self):
        bucket_name_env = f"{self.resource_id}_bucket_name"
        bucket_name = os.getenv(bucket_name_env)
        if bucket_name is None:
            raise ValueError(f"{bucket_name_env} is not defined")
        return bucket_name

    def put_sync(self, key: str, body: Blob):
        return s3().put_object(Bucket=self.bucket_name, Key=key, Body=body)

    @integration("put")
    async def put(self, key: str, body: Blob):
        async with aio_s3() as s3:
            await s3().put_object(Bucket=self.bucket_name, Key=key, Body=body)  # type: ignore

    @integration("get")
    def get_sync(self, key: str) -> Object[StreamingBody]:
        response = s3().get_object(Bucket=self.bucket_name, Key=key)
        return Object(
            bucket=self,
            key=key,
            body=response["Body"],
            etag=response["ETag"],
        )

    @integration("get")
    async def get(self, key: str) -> Object[StreamingBody]:
        async with aio_s3() as s3:
            response = await s3.get_object(Bucket=self.bucket_name, Key=key)
            return Object(
                bucket=self,
                key=key,
                body=response["Body"],
                etag=response["ETag"],
            )

    @integration("delete")
    def delete_sync(self, key: str):
        s3().delete_object(Bucket=self.bucket_name, Key=key)

    @integration("delete")
    async def delete(self, key: str):
        async with aio_s3() as s3:
            await s3.delete_object(Bucket=self.bucket_name, Key=key)

    @integration("list")
    def list_sync(
        self, prefix: str, *, limit: int | None, next_token: str | None = None
    ):
        response = s3().list_objects_v2(
            Bucket=self.bucket_name,
            MaxKeys=limit or 100,
            ContinuationToken=cast(str, next_token or None),
            Prefix=prefix,
        )
        return self.parse_list_objects_v2_response(response)

    @integration("list")
    async def list(
        self, prefix: str, *, limit: int | None, next_token: str | None = None
    ):
        async with aio_s3() as client:
            response = await client.list_objects_v2(
                Bucket=self.bucket_name,
                MaxKeys=limit or 100,
                ContinuationToken=cast(str, next_token or None),
                Prefix=prefix,
            )
            return self.parse_list_objects_v2_response(response)

    def parse_list_objects_v2_response(
        self, response: ListObjectsV2OutputTypeDef
    ) -> ListObjectsResponse:
        return ListObjectsResponse(
            bucket=self,
            objects=[
                ObjectRef(self, obj_ref["Key"], obj_ref["ETag"])  # type: ignore
                for obj_ref in response.get("Contents", [])
            ],
            next_token=response.get("NextContinuationToken"),
        )

    def on(
        self,
        scope: BucketSubscriptionScope | List[BucketSubscriptionScope],
        prefix: Optional[str] = None,
        function_id: Optional[str] = None,
        *,
        with_: Optional[DependencyGroup] = None,
        without: DependencyGroup | None = None,
        # deprecated = None, use with_ and without
        dev: bool | None = None,
        all_extras: bool | None = None,
        without_hashes: bool | None = None,
        without_urls: bool | None = None,
    ):
        def decorate(handler: Callable[[Bucket.ObjectCreatedEvent], Any]):
            from aws_lambda_typing.events.event_bridge import EventBridgeEvent
            from aws_lambda_typing.events.s3 import S3

            # see https://kevinhakanson.com/2022-04-10-python-typings-for-aws-lambda-function-events/
            @function(
                file_name=handler.__code__.co_filename,
                function_id=(
                    function_id
                    if function_id is not None
                    else get_fully_qualified_name(handler)
                ),
                with_=with_,
                without=without,
                dev=dev,
                all_extras=all_extras,
                without_hashes=without_hashes,
                without_urls=without_urls,
            )
            async def lambda_func(event: EventBridgeEvent, context: Any):
                event_detail = S3(event["detail"])  # type: ignore
                result = await handler(
                    Bucket.ObjectCreatedEvent(
                        key=event_detail["object"]["key"],  # type: ignore
                        etag=event_detail["object"]["etag"],  # type: ignore
                        size=event_detail["object"]["size"],  # type: ignore
                    )
                )
                return result

            self.subscriptions.append(
                BucketSubscription(
                    self,
                    scope if isinstance(scope, List) else [scope],
                    lambda_func,
                    prefix,
                )
            )

            return lambda_func

        return decorate

    @integration("get", "list")
    def binaryFiles(
        self, prefix: str | None = None, *, minPartitions: int | None = None
    ) -> RDD[Tuple[str, bytes]]:
        return self.sc.binaryFiles(
            f"s3://{self.bucket_name}/{prefix or ''}", minPartitions=minPartitions
        )



from __future__ import annotations

import json
import os
from typing import Any, Callable, Generic, Type, TypeVar, cast

from pydantic import BaseModel

from packyak.runtime.function import LambdaFunction, function
from packyak.runtime.integration import integration
from packyak.resource import Resource
from packyak.spec import DependencyGroup
from packyak.util.fqn import get_fully_qualified_name
from packyak.util.memoize import memoize
from packyak.util.typed_resource import TypedResource

import boto3
from types_aiobotocore_sqs import SQSClient
from types_aiobotocore_sqs.type_defs import ReceiveMessageResultTypeDef


sqs = memoize(lambda: boto3.client("sqs"))  # type: ignore
session = memoize(lambda: boto3.Session())
aio_sqs = memoize(lambda: TypedResource[SQSClient](session().client("sqs")))  # type: ignore


Body = str | BaseModel

B = TypeVar("B", bound=Body)


class ReceivedMessagesEvent(Generic[B]):
    def __init__(self, messages: list[Message[B]]):
        self.messages = messages


class Message(Generic[B]):
    def __init__(self, queue: "Queue[B]", message_id: str, body: B):
        self.queue = queue
        self.message_id = message_id
        self.body = body

    async def delete(self) -> None:
        await self.queue.delete(receipt_handle=self.message_id)


class QueueSubscription:
    def __init__(
        self,
        queue: "Queue[Any]",
        function: LambdaFunction[Any, Any],
    ):
        self.queue = queue
        self.function = function


class Queue(Generic[B], Resource):
    subscriptions: list[QueueSubscription]

    def __init__(self, resource_id: str, model: Type[B] = str, fifo: bool = False):
        super().__init__(resource_id)
        self.model = model
        self.fifo = fifo
        self.subscriptions = []

    @property
    def queue_url(self):
        queue_url_env = f"{self.resource_id}_queue_url"
        queue_url = os.getenv(queue_url_env)
        if queue_url is None:
            raise ValueError(f"{queue_url_env} is not defined")
        return queue_url

    def send_sync(self, body: Body) -> None:
        sqs().send_message(QueueUrl=self.queue_url, MessageBody=to_json(body))

    @integration("write")
    async def send(self, body: Body) -> None:
        async with aio_sqs() as sqs:
            await sqs.send_message(QueueUrl=self.queue_url, MessageBody=to_json(body))

    def receive_sync(self) -> list[Message[B]]:
        return self._parse_receive(sqs().receive_message(QueueUrl=self.queue_url))

    @integration("read")
    async def receive(self) -> list[Message[B]]:
        async with aio_sqs() as sqs:
            return self._parse_receive(
                await sqs.receive_message(QueueUrl=self.queue_url)
            )

    def _parse_receive(self, response: ReceiveMessageResultTypeDef) -> list[Message[B]]:
        return [
            Message(self, msg["MessageId"], self.from_json(msg["Body"]))  # type: ignore
            for msg in response.get("Messages", [])
        ]

    @integration("delete")
    async def delete(self, receipt_handle: str) -> None:
        async with aio_sqs() as sqs:
            await sqs.delete_message(
                QueueUrl=self.queue_url, ReceiptHandle=receipt_handle
            )

    @integration("list")
    async def list(self):
        async with aio_sqs() as sqs:
            response = await sqs.list_queues()
            return response.get("QueueUrls", [])

    @integration("change_visibility")
    async def change_visibility(self, receipt_handle: str, visibility_timeout: int):
        async with aio_sqs() as sqs:
            await sqs.change_message_visibility(
                QueueUrl=self.queue_url,
                ReceiptHandle=receipt_handle,
                VisibilityTimeout=visibility_timeout,
            )

    def consumer(
        self,
        function_id: str | None = None,
        *,
        with_: DependencyGroup | None = None,
        without: DependencyGroup | None = None,
        # deprecated = None, use with_ and without
        dev: bool | None = None,
        all_extras: bool | None = None,
        without_hashes: bool | None = None,
        without_urls: bool | None = None,
    ):
        def decorate(handler: Callable[[ReceivedMessagesEvent[B]], Any]):
            from aws_lambda_typing.events.sqs import SQSEvent

            # see https://kevinhakanson.com/2022-04-10-python-typings-for-aws-lambda-function-events/
            @function(
                file_name=handler.__code__.co_filename,
                function_id=(
                    function_id
                    if function_id is not None
                    else get_fully_qualified_name(handler)
                ),
                with_=with_,
                without=without,
                dev=dev,
                all_extras=all_extras,
                without_hashes=without_hashes,
                without_urls=without_urls,
            )
            async def lambda_func(event: SQSEvent, context: Any):
                result = await handler(
                    ReceivedMessagesEvent(
                        messages=[
                            Message(
                                queue=self,
                                message_id=message["messageId"],
                                body=self.from_json(message["body"]),
                            )
                            for message in event["Records"]
                        ]
                    )
                )
                # TODO: adapt response
                return result

            self.subscriptions.append(QueueSubscription(self, lambda_func))

        return decorate

    def from_json(self, body: str) -> B:
        if issubclass(self.model, BaseModel):
            return cast(B, self.model.model_validate_json(body))  # type: ignore
        else:
            return json.loads(body)


def to_json(body: Body):
    if isinstance(body, BaseModel):
        return body.model_dump_json()
    else:
        return str(body)



import click
import os
from tabulate import tabulate

from packyak.cli.cli import cli
from packyak.util.emr import EMR


@cli.command()
@click.option(
    "--profile", type=str, help="AWS CLI profile to use when authenticating to SSM"
)
@click.option("-v", "--verbose", type=str, is_flag=True)
def list(profile: str | None, verbose: bool = False):
    if profile is not None:
        os.environ["AWS_PROFILE"] = profile

    emr = EMR()

    try:
        clusters = emr.list_clusters(active_only=True)
        if len(clusters) > 0:
            print(
                tabulate(
                    [
                        [
                            cluster.cluster_id,
                            cluster.cluster_name,
                            cluster.cluster_status.value,
                        ]
                        for cluster in clusters
                    ],
                    # headers=["Cluster ID", "Cluster Name", "Cluster Status"],
                    headers=[],
                    tablefmt="plain",
                )
            )
        else:
            pass
    except Exception as e:
        raise click.ClickException(f"Error listing EMR clusters: {e}")



from typing import Any, Callable


def get_fully_qualified_name(function: Callable[..., Any] | type):
    module_name = function.__module__
    qual_name = function.__qualname__
    return f"{module_name}.{qual_name}"



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple

from packyak.runtime.integration import integration

if TYPE_CHECKING:
    from packyak.storage.bucket import Bucket
    from packyak.spec import BucketSubscriptionScope
    from pyspark import SparkContext, RDD


class Folder:
    sc: SparkContext | None = None

    def __init__(self, parent: "Bucket | Folder", name: str):
        self.parent = parent
        self.name = name

    def __str__(self) -> str:
        return f"{self.bucket}{self.path}/"

    @property
    def resource_id(self) -> str:
        return self.bucket.resource_id

    @property
    def resource_type(self) -> str:
        return self.bucket.resource_type

    @property
    def path(self) -> str:
        from packyak.storage.bucket import Bucket

        if isinstance(self.parent, Bucket):
            return self.name
        else:
            return f"{self.parent.path}/{self.name}"

    @property
    def prefix(self) -> str:
        return f"{self.path}/*"

    @property
    def bucket(self) -> "Bucket":
        from packyak.storage.bucket import Bucket

        if isinstance(self.parent, Bucket):
            return self.parent
        else:
            return self.parent.bucket

    @integration("get", prefix=prefix)
    def get(self, key: str):
        return self.bucket.get(f"{self.name}/{key}")

    @integration("delete", prefix=prefix)
    def delete(self, key: str):
        return self.bucket.delete(f"{self.name}/{key}")

    @integration("put", prefix=prefix)
    def put(self, key: str, body: str):
        return self.bucket.put(f"{self.name}/{key}", body)

    @integration("list", prefix=prefix)
    async def list(self, prefix: str, *, limit: int | None, next_token: str | None):
        return self.bucket.list(
            f"{self.path}/{prefix}", limit=limit, next_token=next_token
        )

    def folder(self, path: str) -> "Folder":
        return self / path

    # Overload the '/' operator
    def __truediv__(self, other: str) -> "Folder":
        return Folder(self, other)

    def on(self, scope: BucketSubscriptionScope, prefix: str | None = None):
        return self.bucket.on(
            scope, prefix=self.path + prefix if prefix else self.path + "/*"
        )

    @integration("get", "list")
    def binaryFiles(
        self, prefix: str | None = None, *, minPartitions: int | None = None
    ) -> RDD[Tuple[str, bytes]]:
        return self.parent.binaryFiles(prefix, minPartitions=minPartitions)



from enum import Enum
from typing import Callable, TypeVar, TYPE_CHECKING

from packyak.runtime.job import Job
from packyak.spec import DependencyGroup
from packyak.resource import Resource
from pyspark import SparkContext

Return = TypeVar("Return")


class Engine(Enum):
    Spark = "SPARK"


class Cluster(Resource):
    def __init__(self, cluster_id: str):
        super().__init__(cluster_id)
        self.cluster_id = cluster_id
        self.engine = Engine.Spark

    def job(
        self,
        job_id: str | None = None,
        *,
        file_name: str | None = None,
        with_: DependencyGroup = None,
        without: DependencyGroup = None,
        # deprecated, use with_ and without
        dev: bool | None = None,
        all_extras: bool | None = None,
        without_hashes: bool | None = None,
        without_urls: bool | None = None,
    ):
        def decorator(handler: Callable[[SparkContext], Return]) -> Job[Return]:
            return Job(
                handler=handler,
                job_id=job_id,
                file_name=file_name,
                with_=with_,
                without=without,
                dev=dev,
                all_extras=all_extras,
                without_hashes=without_hashes,
                without_urls=without_urls,
            )

        return decorator



from typing import Any
import boto3
from dataclasses import dataclass
from enum import Enum


class ClusterStatus(Enum):
    STARTING = "STARTING"
    BOOTSTRAPPING = "BOOTSTRAPPING"
    RUNNING = "RUNNING"
    WAITING = "WAITING"
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"
    TERMINATED_WITH_ERRORS = "TERMINATED_WITH_ERRORS"


@dataclass
class Cluster:
    cluster_id: str
    cluster_name: str
    cluster_status: ClusterStatus


class EMR:
    def __init__(self):
        self.emr = boto3.client("emr")

    def get_primary_node_instance_id(self, cluster_id: str) -> str:
        instance_id = self.try_get_primary_node_instance_id(cluster_id)
        if instance_id is None:
            raise Exception(f"No primary instance found for EMR cluster {cluster_id}")
        return instance_id

    def try_get_primary_node_instance_id(self, cluster_id: str) -> str | None:
        for group in self.emr.list_instance_groups(ClusterId=cluster_id)[
            "InstanceGroups"
        ]:
            if "InstanceGroupType" not in group:
                raise Exception("No instance group type")
            if group["InstanceGroupType"] == "MASTER":
                if "Id" not in group:
                    raise Exception("No instance group ID")
                instances = self.emr.list_instances(
                    ClusterId=cluster_id, InstanceGroupId=group["Id"]
                )["Instances"]
                if len(instances) > 0:
                    if "Ec2InstanceId" not in instances[0]:
                        raise Exception("No EC2 instance ID")
                    return instances[0]["Ec2InstanceId"]

        return None

    def list_clusters(
        self,
        active_only: bool = True,
    ):
        """
        List EMR clusters based on their states.

        :param states: A list of states to filter the clusters by. If None, all active clusters are listed.
        :return: A list of Cluster objects representing the EMR clusters.
        """

        states = (
            [
                ClusterStatus.STARTING,
                ClusterStatus.BOOTSTRAPPING,
                ClusterStatus.RUNNING,
                ClusterStatus.WAITING,
            ]
            if active_only
            else [
                ClusterStatus.STARTING,
                ClusterStatus.BOOTSTRAPPING,
                ClusterStatus.RUNNING,
                ClusterStatus.WAITING,
                ClusterStatus.TERMINATING,
                ClusterStatus.TERMINATED,
                ClusterStatus.TERMINATED_WITH_ERRORS,
            ]
        )

        response = self.emr.list_clusters(
            ClusterStates=[state.value for state in states]
        )
        clusters = response.get("Clusters", [])

        def parse_cluster_status(cluster: Any):
            return ClusterStatus(cluster["Status"]["State"])

        def fail(msg: str):
            raise Exception(msg)

        return [
            Cluster(
                cluster_id=cluster["Id"] if "Id" in cluster else fail("No cluster ID"),
                cluster_name=cluster["Name"]
                if "Name" in cluster
                else fail("No cluster name"),
                cluster_status=parse_cluster_status(cluster),
            )
            for cluster in clusters
        ]



from typing import (
    Callable,
    TypeVar,
    TYPE_CHECKING,
)
from packyak.runtime.runnable import Runnable

from packyak.spec import DependencyGroup


# if TYPE_CHECKING:
from pyspark import SparkContext

Return = TypeVar("Return")


class Job(Runnable[[SparkContext], Return]):
    def __init__(
        self,
        handler: Callable[[SparkContext], Return],
        job_id: str | None = None,
        file_name: str | None = None,
        with_: DependencyGroup | None = None,
        without: DependencyGroup | None = None,
        dev: bool | None = None,
        all_extras: bool | None = None,
        without_hashes: bool | None = None,
        without_urls: bool | None = None,
    ) -> None:
        super().__init__(
            resource_id=job_id,
            handler=handler,
            file_name=file_name,
            with_=with_,
            without=without,
            dev=dev,
            all_extras=all_extras,
            without_hashes=without_hashes,
            without_urls=without_urls,
        )
        self.job_id = job_id



from pydantic import Field
from typing import Any, NamedTuple, TypeVar


# Define a generic type variable
PartitionKeyValue = str | int | float | bool | None

T = TypeVar("T", bound=PartitionKeyValue)


def PartitionKey() -> Any:
    return Field(..., partition_key=True)  # type: ignore - i think pydantic actually allows this


class PartitionKeyField(NamedTuple):
    name: str
    type: type[PartitionKeyValue]



from packyak.runtime.function import LambdaFunction
from packyak.runtime.job import Job
from packyak.duration import TimeUnit
from typing import Any, TypeVar

T = TypeVar("T", bound=Job[Any] | LambdaFunction[[], Any])


def cron(cron: str):
    def wrapper(func: T) -> T:
        return func

    return wrapper



import os
from typing import Any

from packyak.synth.synth import synth


async def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--root", type=str, help="root directory", required=True)
    args = parser.parse_args()

    packyak_spec = await synth(root_dir=args.root)
    packyak_dir = ".packyak"

    if not os.path.exists(packyak_dir):
        os.makedirs(packyak_dir)
    with open(f"{packyak_dir}/spec.json", "w") as f:
        spec = packyak_spec.model_dump(
            exclude_unset=True,
            exclude_none=True,
        )

        def transform(obj: Any) -> Any:
            if isinstance(obj, dict):
                d = dict(
                    (
                        k,
                        transform(v),
                        # TODO: decide whether the file_name should be relative or absolute
                        # os.path.relpath(v, packyak_dir)
                        # if k == "file_name"
                        # else transform(v),
                    )
                    for k, v in obj.items()
                    if not k.startswith("_")
                )
                return d

            elif isinstance(obj, list):
                return [transform(v) for v in obj]
            return obj

        import json

        f.write(json.dumps(transform(spec), indent=2))


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())
